<?php
/*
+-----------------------------------------------------------------------------+
| (c) Copyright IBM Corporation 2006, 2007.                                   |
| All Rights Reserved.                                                        |
+-----------------------------------------------------------------------------+
| Licensed under the Apache License, Version 2.0 (the "License"); you may not |
| use this file except in compliance with the License. You may obtain a copy  |
| of the License at -                                                         |
|                                                                             |
|                   http://www.apache.org/licenses/LICENSE-2.0                |
|                                                                             |
| Unless required by applicable law or agreed to in writing, software         |
| distributed under the License is distributed on an "AS IS" BASIS, WITHOUT   |
| WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.            |
| See the License for the specific language governing  permissions and        |
| limitations under the License.                                              |
+-----------------------------------------------------------------------------+
| Author: Wangkai Zai                                                         |
|                                                                             |
+-----------------------------------------------------------------------------+
*/
if ( ! class_exists('SCA_Bindings_message_SAMClient', false) ) {
    class SCA_Bindings_message_SAMClient {
    
        private $callback = null;
        private $connection = null;
        private $queuename = null;  //the primary queue

        /**
         * Constructor - 
         * @param $callback a object instance of either the proxy or the wrapper
         *               which should implement an onMessage method
         */
        public function __construct($callback)
        {
             $this->callback = $callback;
             $this->connection = new SAMConnection();
        }


       /**
        * function that listens to a specified queue
        * and calls ($this->callback)'s onMessage method if message received 
        * 
        * @param $queue string URL of the traget queue
        */
        public function start($queue = null)
        {
            SCA::$logger->log('entering');

            // if target queue not specified, the primary queue will be used
            if(is_null($queue)){
                $queue = $this->queuename;
            }
            echo "Listener for $queue has been started. To exit, press <Ctrl> + <C> ...\n";

            // to avoid the 30 seconds PHP maximun execution time error
            set_time_limit(0);
            while(1){
                //$conn->receive($queuename, array(SAM_WAIT=>10000, SAM_CALLBACKFUNC=>"displaymessage"));
                $msg = $this->connection->receive($queue, array(SAM_WAIT=>0));
                if($msg){
                    $response = $this->callback->onMessage($msg) ;

                    /*check the target queue for callbacks using the 'scaCallbackQueue' user property 
                      or the JMS replyTo header*/
                    $response_queue = false;
                    if (isset($msg->header->scaCallbackQueue)) {
                        $response_queue = trim( $msg->header->scaCallbackQueue );
                    }else if (isset($msg->header->SAM_REPLY_TO)) {
                        $response_queue = trim( $msg->header->SAM_REPLY_TO );
                    }

                    /*if callback is defined send response to the queue*/
                    if($response_queue){
                        $response_body = isset($response)? $response: ''; 
                        $response_msg = new SAMMessage($response_body);
                        $response_msg->header->SAM_TYPE = SAM_TEXT;

                        if (isset($msg->header->SAM_CORRELID)) {
                            $correlid = $msg->header->SAM_CORRELID ;
                        }else{
                            $correlid = $msg->header->SAM_MESSAGEID;
                        }
                        SCA::$logger->log("sending response $response_body to queue $response_queue");
                        $this->msgput($response_msg,$correlid, $response_queue);
                    }
                }
            }

            SCA::$logger->log("exiting");
            return;
        }

        /**
         * function that send message to a borker
         * @param $msg SAMMessage object to be sent
         * @param $corrid string criteria for message selection
         * @param $queue string URL of the traget queue
         */
        public function msgput( $msg, $corrid = null, $queue = null){
            SCA::$logger->log('entering');

            // if target queue not specified, the primary queue will be used
            if(is_null($queue)){
                $queue = $this->queuename;
            }

            $options = array() ;
            if (!is_null($corrid)) {
                $options[SAM_CORRELID] = $corrid;
            }

            SCA::$logger->log("Jack: sending msg to queue $queue");
            $msgid = $this->connection->send($queue, $msg, $options);
            if(!$msgid){
                SCA::$logger->log('SAM: send message failed: ('.$this->connection->errno.') '.$this->connection->error);
            }
            return $msgid;
        }

        /**
         * function that send message from a borker
         * @param $corrid string criteria used to message selection
         * @param $queue string URL of the traget queue
         * @param $timeout integer how long should wait 
         */
        public function msgget($corrid = null, $queue = null, $timeout = 5000){
            SCA::$logger->log('entering');

            // if target queue not specified, the primary queue will be used
            if(is_null($queue)){
                $queue = $this->queuename;
            }

            $options[SAM_WAIT] = $timeout ;
            if (!is_null($corrid)) {
                $options[SAM_CORRELID] = $corrid;
            }

            $msg = $this->connection->receive($queue, $options);
            return $msg;
        }

        /**
         * Function that apply binding configuration
         * and creates a SAMConnection
         * 
         * @param $config array OR object 
         */
        public function create_connection($config){
            SCA::$logger->log("entering");
            
            /*applying binding configuration*/;
            $optionsarray = array();
            foreach ($config as $key => $value){
                switch ($key) {
                case "protocol":
                    $protocol = $value ;
                    break;
                case "queueurl":
                    $this->queuename = $value ;
                    break;
                case "host":
                    $optionsarray[SAM_HOST] = $value ;
                    break;
                case "port":
                    $optionsarray[SAM_PORT] = $value ;
                    break;
                case "broker":
                    $optionsarray[SAM_BROKER] = $value ;
                    break;
                case "endpoints":
                    $optionsarray[SAM_ENDPOINTS] = $value ;
                    break;
                case "targetchain":
                    $optionsarray[SAM_TARGETCHAIN] = $value ;
                    break;
                case "bus":
                    $optionsarray[SAM_BUS] = $value ;
                    break;
                case "userid":
                    $optionsarray[SAM_USERID] = $value ;
                    break;
                case "password":
                    $optionsarray[SAM_PASSWORD] = $value ;
                    break;
                }
            }

            /*check mandatory fields */
            if (!isset($protocol)) {
                throw new SCA_RuntimeException('"protocol" is mandatory for MS binding. ');
            }
            if (is_null($this->queuename)) {
                throw new SCA_RuntimeException('"queueurl" is mandatory for MS binding. ');
            }

            $this->connection->connect($protocol, $optionsarray);
            if (!$this->connection) {
                throw new SCA_RuntimeException("SAM connection failed");
            }
            SCA::$logger->log("exiting");
        }

        public function getLastError(){
            if ($this->connection->errno) {
                return '('.$this->connection->errno.') '.$this->connection->error;
            }else{
                return 'There was no error occured.';
            }
        }


        public function disconnect(){
            $this->connection->disconnect();
        }
    }/* End SCA_Bindings_message_SAMClient class*/
}/* End instance check*/
?>
